; FLOW CONTROL

; Satements vs. Expressions

; In Java, expressions return values whereas statements do not. Howver, in
; Clojure everything is an expression so everything returns a value. Blocks of
; multiple expressions return the last value and expressions that exclusively
; have side effects return nil.

; Accordingly, flow control operators are expressions, can be composed, and
; return their result.

; if consists of a condition, a "then", and an "else".
; the else is optional

(str "2 is " (if (even? 2) "even" "odd"))
; 2 is even
(if (true? false) "impossible!")
; nil

; In Clojure, all values are logically true or false. The only "false" variables
; are false and nil.

(if true :truthy :falsey)
; :truthy
(if (Object.) :truthy :falsey) ; objects are true
; :truthy
(if [] :truthy :falsey) ; empty collections are true
; :truthy
(if 0 :truthy :falsey) ; zero is true
; :truthy
(if false :truthy :falsey)
; :falsey
(if nil :truthy :falsey)
; :falsey

; If only takes a single expression for the "then" and "else". Use do to create
; larger blocks that are a single expression. The only reason to do this is if
; your bodies have side-effects

(if (even? 5)
  (do (println "even")
      true)
  (do (println "odd")
      false))

; when is an if with only a then branch, but it evaluates any number of
; statements as a body so no do is required

(when (neg? x)
  (throw (RuntimeException. (str "x must be positive: " x))))

; cond is a series of tests and expressions. Each test is evaluated in order and
; the expression is evaluated and returned for the first true test.

; if no test is satisfied, nil is returned. A common idiom is to use a final
; test of :else (which always evaluates to true)

(let [x 11]
  (cond
        (< x 2) "x is less than 2"
        (< x 10) "x is less than 10"
        :else "x is greater than or equal to 10"))

; case compares an argument to a series of values to find a match. This is done
; in constant time. However, each value must be a compile-time literal (numbers,
; strings, keywords, etc.). Unlike cond, case will throw an exception if no
; value matches.

(let [x 11]
  (case
    5  "x is 5"
    10 "x is 10"))
; IllegalArgumentException No matching clause: 11

; case can have a final trailing expression that will be evaluated if no test
; matches

(let [x 11]
  (case
    5  "x is 5"
    10 "x is 10"
    "x isn't 5 or 10"))
; x isn't 5 or 10

; ITERATION FOR SIDE EFFECTS

; dotimes evaluates an expression n times then and returns nil

(dotimes [i 3]
  (println i))
; 0
; 1
; 2
; nil

; doseq iterates over a sequence, forces evaluation of the sequence if it's
; lazy, and then returns nil

(doseq [n (range 3)]
         (println n))
; 0
; 1
; 2
; nil

; doseq can have multiple bindings, similar to nested foreach loops. It
; processes all permutations of sequence content then returns nil.

(doseq [letter [:a :b]
               number (range 3)] ; list of 0, 1, 2
         (prn [letter number]))
; [:a 0]
; [:a 1]
; [:a 2]
; [:b 0]
; [:b 1]
; [:b 2]
; nil

; CLOJURE'S FOR

; It is a list comprehension, not a for loop, basically a generator function for
; sequence permutation
; Bindings behave like doseq
(for [letter [:a :b]
             number (range 3)] ; list of 0, 1, 2
         [letter number])
; ([:a 0] [:a 1] [:a 2] [:b 0] [:b 1] [:b 2])

; RECURSION

; Clojure provides recur and the sequence abstraction
; recur is classic recursion (consumers don't control it, considered low-level)
; Sequences represent iteration as values (consumers can partially iterate)
; Reducers represent iteration as function composition (added in Clojure 1.5)

; loop and recur are a functional looping construct where loop defines bindings
; and recur executes loop with new bindings.

(loop [i 0]
  (if (< i 10)
    (recur (inc i))
    i))

; defn and recur -> function arguments are implicit loop bindings

(defn increase [i]
  (if (< i 10)
    (recur (inc i))
    i))

; When using recur for recursion, recur must be in "tail position" (the last
; expression in a branch) and it must provide values for all bound symbols by
; position (i.e. can't call recur without passing values back to loop or defn/fn
; arguments). Recursion via recur does not consume stack.

; EXCEPTIONS

; Exception handling is done by throw/catch/finally as in Java:

(try
  (/ 2 1)
  (catch ArithmeticException e
    "divide by zero")
  (finally
    (println "cleanup")))

; Exceptions can also be thrown:

(try
  (throw (Exception. "something went wrong"))
  (catch Exception e (.getMessage e)))

; Exceptions with Clojure data:

; ex-info takes a message and a map
; ex-data gets the map back out or nil of not created with ex-info

(try
  (throw (ex-info "There was a problem" {:detail 42}))
  (catch Exception e
    (prn (:detail (ex-data e)))))

; Clojure has a with-open function similar to Python's "with open("file") as f":

(let [f (clojure.java.io/writer "/tmp/new")]
  (try
    (.write f "some text")
    (finally
      (.close f))))

; Can be written:
(with-open [f (clojure.java.io/writer "/tmp/new")]
  (.write f "some text"))
