; HASHED COLLECTIONS

; Sets and maps are hashed collections designed for efficient element lookup

; SETS

; Like mathematical sets: unordered and without duplicates
; Efficient for checking whether a collection contains an element or removing an
; arbitrary element

(def players #{"Alice", "Bob", "Kelly"})

; conj can be used to add elements:

(conj players "Fred")
;#{"Alice" "Fred" "Bob" "Kelly"}

; disj (disjoin) can be used to remove one or more elements from a set:
(disj players "Bob" "Sal")
;#{"Alice" "Fred" "Kelly"}

; It's okay to disj elements that don't exist in the set

; Check if a set contains an element like so:

(contains? players "Kelly")
; true

; Sorted sets are sorted according to a comparator function which can compare
; two elements. By default, Clojure's compare function is used, which sorts in
; "natural" order for numbers, strings, etc.

(conj (sorted-set) "Bravo" "Charlie" "Sigma" "Alpha")
; #{"Alpha" "Bravo" "Charlie" "Sigma"}

; A custom comparator can also be used with sorted-set-by

; into is used for putting one collection into another. It returns a collection
; of the same type as its first argument

(def players #{"Alice" "Bob" "Kelly"})
(def new-players ["Tim" "Sue" "Greg"])
(into players new-players)
;#{"Alice" "Greg" "Sue" "Bob" "Tim" "Kelly"}

; MAPS

; Commonly used to: manage an association of keys to values and to represent
; domain application data

; Literal maps
(def scores {"Fred" 1400
             "Bob"  1240
             "Angela" 1024})

; Note that commas are treated as whitespace in Clojure, they can be used as
; needed for human readability

; Add new values to maps with assoc (associate). If the key already exists, the
; value is replaced (updated):

(assoc scores "Sally" 0)
; {"Angela" 1024, "Bob" 1240, "Fred" 1400, "Sally" 0}

; Remove key-value pairs with dissoc (dissociate):

(dissoc scores "Bob")
; {"Angela" 1024, "Fred" 1400}

; Look up a value with get:
(get scores "Angela")
; 1024

; Or when the map in question is treated as a constant lookup table, it can be
; invoked itself:
(def directions {:north 0
                 :east 1
                 :south 2
                 :west 3})
(directions :north)
; 0

; You can also look an item up but fall back to a default value if it's not
; found like so:
(get scores "Sam" 0)
; 0
(directions :northwest -1)
; -1

; This is useful for distinguishing between a missing key and an existing key
; with a 'nil' value

; There are two ways of checking whether a map contains an entry:

(contains? scores "Fred")
; true

(find scores "Fred")
["Fred" 1400]

; You can also just get the keys or the values in a map:

(keys scores)
; ("Fred" "Bob" Angela")
(vals scores)
; (1400 1240 1024)

; The zipmap function can be used to "zip" together two sequences into a map:

(def players #{"Alice" "Bob" "Kelly"})
(zipmap players (repeat 0))
; {"Kelly" 0, "Bob" 0, "Alice" 0}

; There are more ways using Clojure's sequence functions, but this will be
; touched on later:

; with map and into
(into {} (map (fn [player] [player 0]) players))
; with reduce
(reduce (fn [m player]
          (assoc m player 0))
        {} ; initial value
        players)

; The merge function can be used to combine multiple maps into a single map. If
; they both contain the same key, the rightmost one wins.

(def new-scores {"Angela" 300 "Jeff" 900})
(merge scores new-scores)
;{"Fred" 1400, "Bob" 1240, "Jeff" 900, "Angela" 300}

; Alternately, use merge-with to supply a function to invoke when there is a
; conflict:

(def new-scores {"Fred" 550 "Angela" 900 "Sam" 1000})
(merge-with + scores new-scores)
;{"Sam" 1000, "Fred" 1950, "Bob" 1240, "Angela" 1924}

; Similar to sorted sets, sorted maps maintain the keys in sorted order based on
; a comparator, using compare as the default comparator function

(def sm (sorted-map
          "Bravo" 204
          "Alfa" 35
          "Sigma" 99
          "Charlie" 100))
;{"Alfa" 35, "Bravo" 204, "Charlie" 100, "Sigma" 99}
(keys sm)
;("Alfa" "Bravo" "Charlie" "Sigma")
(vals sm)
;(35 204 100 99)

; REPRESENTING APPLICATION DOMAIN INFORMATION

; When needing to represent a lot of domain information with the same set of
; fields known in advance, you can use a map with keyword keys:
(def person
  {:first-name "Kelly"
   :last-name "Keen"
   :age 32
   :occupation "Programmer"})

; Which still lets you do things such as:

(get person :occupation)
; "Programmer"

(person :occupation)
; "Programmer"

; However, the most common way to get field values for this isue is like so:

(:occupation person)
; "Programmer"

; or, with a default value:

(:favorite-color person "beige")
; "beige"

; assoc and dissoc work the same as well:

(assoc person :occupation "Baker")
; {:age 32, :last-name "Keen", :first-name "Kelly", :occupation "Baker"}

(dissoc person :age)
; {:last-name "Keen", :first-name "Kelly", :occupation "Programmer"}

; It's common to see entities nested inside of other entities:

(def company
  {:name "WidgetCo"
   :address {:street "123 Main St"
             :city "Springfield"
             :state "IL"}})

; In these cases, get-in can be used to access fields at any level inside:

(get-in company [:address :city])
; "Springfield"

; assoc-in and update-in work similarly as well:
(assoc-in company [:address :street] "303 Broadway")
; {:name "WidgetCo",
;  :address
;  {:state "IL",
;   :city "Springfield",
;   :street "303 Broadway"}}

; RECORDS

; An alternative to using maps is to create a record. They are designed
; specifically for this use case and generally have better performance. They
; also have a named "type" which can be used for polymorphic behaviour.

; Records are defined with the list of field names for record instances which
; will be treated as keyword keys in each record instance:

;; Define a record structure
(defrecord Person [first-name last-name age occupation])

;; Positional constructor - generated
(def kelly (->Person "Kelly" "Keen" 32 "Programmer"))

;; Map constructor - generated
(def kelly (map->Person
             {:first-name "Kelly"
              :last-name "Keen"
              :age 32
              :occupation "Programmer"}))

; Records are used almost exactly the same as maps, but cannot be invoked as a
; function like maps:

user=> (:occupation kelly)
; "Programmer"
