; NAMESPACES

; Namespaces provide a means to organize code. They allow giving unambiguous
; names to functions or values. These full names are long, thus namespaces also
; provide a means to unambiguously reference the names of other functions and
; values but using names that are shorter and easier to type.

; A namespace is both a name context and a container for vars. Namespace
; names are symbols where periods are used to separate namespace parts, such as
; clojure.string. By convention, namespace names are typically lowercase and use
; - to separate words, though this is not required.

; VARS

; vars in a namespace have a fully-qualified name that is the combination of the
; namespace and the var name. E.g. clojure.string/join is a fully-qualified var
; name where clojure.string is the namespace and join is the var inside. All
; vars are globally accessible by their fully-qualified name. By convention,
; var names are lowercase with words separated by -, though this is also not
; required. Var names may contain most non-whitespace characters.

; vars are created using def and other special forms or macros that start with
; def such as defn. They are created in the current namespace, which Clojure
; tracks in the var clojure.core/*ns*. The current namespace can be changed with
; the in-ns function.

; LOADING

; In addition, namespace names provide a convention for where a namespace's code
; should be found for loading. A path is created based on the namespace's name.

; Periods become directory separators
; Hyphens become underscores
; The file extension .clj is added

; Thus, the namespace com.some-example.my-app becomes
; com/some_example/my_app.clj.

; Load paths are searched using the JVM classpath, which is a series of
; directory locations or JAR files (which are essentially just zip files).

; When a resource is needed, the JVM searches each classpath location in order
; for a file at the relative location of the load path. If the classpath was
; src:test, the load path would be checked at src/com/some_example/my_app.clj
; then test/com/some_example/my_app.clj.

; There are several ways to load code in Clojure, but most commonly it's done
; using require.

; Due to this loading convention, most Clojure is structured with a 1-to-1
; mapping of namespaces to files, stored in hierarchical fashion that maps to
; the namespace structure.

; DECLARING NAMESPACES

; Most Clojure files represent a single namespace and declare the dependencies
; for that namespace at the top of the file using the ns macro which often looks
; like this:

(ns com.some-example.my-app
  "My app example"
  (:require
    [clojure.set :as set]
    [clojure.string :as str]))

; The ns macro specifies the namespace name (should match the file path location
; using the conventions above), an optional docstring, and then one or more
; clauses that declare things about the namespace.

; By default, we can refer to or invoke vars in the current namespace without
; specifying the namespace. We can also usually refer to clojure.core library
; functions without fully qualifying them either. All clojure.core library vars
; have been referred into the current namespace. refer makes an entry in the
; current namespace's symbol table that refers to the var in the other
; namespace.

; The clojure.core referral is done by the ns macro, and there are ways to
; suppress this if needed.

; The :require clause corresponds to the require function which specifies one or
; more namespaces to load that this namespace depends on. For each namespace,
; require can do several things:

; Load or reload the namespace
; Optionally assign an alias that can be used to refer to vars from the loaded
;   namespace in this scope
; Optionally refer vars from the loaded namespace for use by unqualified name in
;   this namespace.

; The last two parts are all about making names easier to use. While vars can
; always be referred to by their fully-qualified name, we rarely want to type
; fully-qualified names in our code. Aliases let us use shorter versions and
; refer allows us to use names without a namespace qualifier at all.

; In require, namespaces most commonly take one of three forms:

; clojure.set (just loads clojure.set namespace if not already loaded)
; [clojure.set :as set] (load and create an alias set for the whole clojure.set
;   namespace)
; [clojure.set :refer [union intersection]] (load and refer specific vars into
;   this namespace)

; This is similar to Python's:

; import sys
; import sys as s
; from sys import argv

; JAVA CLASSES AND IMPORTS

; In addition to vars, Clojure also provides support for Java interop and access
; to Java classes which live in packages. Java classes can always be referred to
; using their FQN, e.g. java.util.Date.

; The ns macro also imports the classes in the java.lang package so that they
; can be used as just the class name, rather than the fully-qualified class
; name. e.g. String instead of java.lang.String.

; Similar to :refer, the ns macro has an :import clause (supported by the import
; macro) that lets you import other classes so they can be used with unqualified
; names:

(ns com.some-example.my-app2
  (:import
    [java.util Date UUID]
	[java.io File]))

; This example imports the Date and UUID class from the java.util package and
; the File class from the java.io package.
