(ns text-editor.core
  (:use seesaw.core
        seesaw.mig
        seesaw.chooser
        [clojure.java.io :only [file]]))

; Make the window look like a native app
(native!)

(def current-file
  (atom (file "")))

(def current-file-label
  (if (.exists @current-file)
    (label :text @current-file)
    (label :text "No file selected")))

(def status-label
  (label :text "Welcome!"))

(defn set-status [& strings]
  (text! status-label (apply str strings)))

(def editor
  (if (.exists @current-file)
    (editor-pane :id "editor" :text (slurp @current-file))
    (editor-pane :id "editor" :text "")))

(defn calculate-linecount []
   (if (.exists @current-file)
     (->> (slurp @current-file)
          (re-seq #"\n")
          count)
    1))

(defn format-linecount [count]
  (vec (map str (range 1 (inc count)))))

(def linecount
  (vertical-panel
    :items (format-linecount (calculate-linecount))))

(defn set-current-file [f]
  (swap! current-file (constantly f)))

(defn select-file [type]
  (choose-file editor :type type))

(defn do-new [event]
  (set-current-file (file ""))
  (text! editor "")
  (set-status "New file"))

(def new-action
  (action
    :handler do-new
    :name "New"
    :key "menu N"
    :tip "Open a new file."))

(defn do-open [event]
  (let [selected (select-file :open)]
    (set-current-file selected))
  (text! editor (slurp @current-file))
  (set-status "File opened"))

(def open-action
  (action
    :handler do-open
    :name "Open"
    :key "menu O"
    :tip "Open a file."))

(defn do-save-as [event]
  (when-let [selected (select-file :save)]
    (set-current-file selected)
    (spit @current-file (text editor))
    (set-status "File saved")))

(defn do-save [event]
  (if-not (.exists @current-file)
    (do-save-as event)
    (do
      (spit @current-file (text editor))
      (set-status "File saved"))))

(def save-action
  (action
    :handler do-save
    :name "Save"
    :key "menu S"
    :tip "Save the document."))

(def save-as-action
  (action
    :handler do-save-as
    :name "Save As"
    :key "menu shift S"
    :tip "Save the document with a different name. "))

(def quit-action
  (action
    :handler (fn [event] (dispose! event))
    :name "Quit"
    :key "menu Q"
    :tip "Exit the program."))

(def copy-action
  (action
    :handler (fn [event] (.copy editor))
    :name "Copy"
    :key "menu C"
    :tip "Copy selected text to the clipboard."))

(def cut-action
  (action
    :handler (fn [event] (.cut editor))
    :name "Cut"
    :key "menu X"
    :tip "Cut selected text from the document into the clipboard."))

(def paste-action
  (action
    :handler (fn [event] (.paste editor))
    :name "Paste"
    :key "menu V"
    :tip "Paste text from the clipboard into the document."))

(def about-action
  (action
    :handler (fn [event] (alert "Text Editor v0.1.0-SNAPSHOT"))
    :name "About"
    :key "menu A"
    :tip "More information about this program."))

(def topbar
  (menubar :items
           [(menu :text "File" :items [new-action open-action save-action save-as-action quit-action])
           (menu :text "Edit" :items [copy-action cut-action paste-action])
           (menu :text "Help" :items [about-action])]))

(def main-panel
  (mig-panel
    :constraints ["fill, ins 0"]
    :items [[(scrollable editor) "grow"]
            [linecount "west, grow"]
            [current-file-label "south"]
            [(separator) "south"]
            [status-label "south"]]))

(defn filemenu []
  (value (select menubar [:file])
  println "File clicked!"
  println value))

(defn -main [& args]
  (add-watch
    current-file
    nil
    (fn [key ref old new]
      (do
        (text! current-file-label (str new))
        (replace! main-panel linecount (vertical-panel :items (format-linecount (calculate-linecount)))))))
  (invoke-later
    (-> (frame :title "Text Editor",
               :menubar topbar
               :content main-panel
               :minimum-size [800 :by 600]
               :on-close :exit)
        pack!
        show!)))
